-- luacheck: globals app os verboseLevel connect
local app = app
local Class = require "Base.Class"
local Unit = require "Unit"
local Fader = require "Unit.ViewControl.Fader"
local GainBias = require "Unit.ViewControl.GainBias"
local InputGate = require "Unit.ViewControl.InputGate"
local OutputScope = require "Unit.ViewControl.OutputScope"
local Encoder = require "Encoder"
local ply = app.SECTION_PLY

local ADSR = Class{}
ADSR:include(Unit)

function ADSR:init(args)
  args.title = "ADSR"
  args.mnemonic = "En"
  Unit.init(self,args)
end

function ADSR:onLoadGraph(channelCount)
  if channelCount==2 then
    self:loadStereoGraph()
  else
    self:loadMonoGraph()
  end
end

function ADSR:loadMonoGraph()
  local gate = self:createObject("Comparator","gate")
  gate:setGateMode()
  local adsr = self:createObject("ADSR","adsr")
  local attack = self:createObject("GainBias","attack")
  local decay = self:createObject("GainBias","decay")
  local sustain = self:createObject("GainBias","sustain")
  local release = self:createObject("GainBias","release")
  local attackRange = self:createObject("MinMax","attackRange")
  local decayRange = self:createObject("MinMax","decayRange")
  local sustainRange = self:createObject("MinMax","sustainRange")
  local releaseRange = self:createObject("MinMax","releaseRange")

  connect(self,"In1",gate,"In")
  connect(gate,"Out",adsr,"Gate")
  connect(adsr,"Out",self,"Out1")

  connect(attack,"Out",adsr,"Attack")
  connect(decay,"Out",adsr,"Decay")
  connect(sustain,"Out",adsr,"Sustain")
  connect(release,"Out",adsr,"Release")

  connect(attack,"Out",attackRange,"In")
  connect(decay,"Out",decayRange,"In")
  connect(sustain,"Out",sustainRange,"In")
  connect(release,"Out",releaseRange,"In")

  self:createMonoBranch("attack",attack,"In",attack,"Out")
  self:createMonoBranch("decay",decay,"In",decay,"Out")
  self:createMonoBranch("sustain",sustain,"In",sustain,"Out")
  self:createMonoBranch("release",release,"In",release,"Out")
end

function ADSR:loadStereoGraph()
  self:loadMonoGraph()
  connect(self.objects.adsr,"Out",self,"Out2")
end


local views = {
  expanded = {"input","attack","decay","sustain","release"},
  collapsed = {},
  input = {"scope","input"},
  attack = {"scope","attack"},
  decay = {"scope","decay"},
  sustain = {"scope","sustain"},
  release = {"scope","release"},
}

function ADSR:onLoadViews(objects,branches)
  local controls = {}

  controls.scope = OutputScope {
    monitor = self,
    width = 4*ply,
  }

  controls.input = InputGate {
    button = "input",
    description = "Unit Input",
    comparator = objects.gate,
  }

  controls.attack = GainBias {
    button = "A",
    branch = branches.attack,
    description = "Attack",
    gainbias = objects.attack,
    range = objects.attackRange,
    biasMap = Encoder.getMap("ADSR"),
    biasUnits = app.unitSecs,
    initialBias = 0.050
  }

  controls.decay = GainBias {
    button = "D",
    branch = branches.decay,
    description = "Decay",
    gainbias = objects.decay,
    range = objects.decayRange,
    biasMap = Encoder.getMap("ADSR"),
    biasUnits = app.unitSecs,
    initialBias = 0.050
  }

  controls.sustain = GainBias {
    button = "S",
    branch = branches.sustain,
    description = "Sustain",
    gainbias = objects.sustain,
    range = objects.sustainRange,
    biasMap = Encoder.getMap("unit"),
    biasUnits = app.unitNone,
    initialBias = 1
  }

  controls.release = GainBias {
    button = "R",
    branch = branches.release,
    description = "Release",
    gainbias = objects.release,
    range = objects.releaseRange,
    biasMap = Encoder.getMap("ADSR"),
    biasUnits = app.unitSecs,
    initialBias = 0.100
  }

  return controls, views
end

return ADSR
