-- luacheck: globals app os verboseLevel connect
local app = app
local Class = require "Base.Class"
local Unit = require "Unit"
local GainBias = require "Unit.ViewControl.GainBias"
local Encoder = require "Encoder"
local ply = app.SECTION_PLY

local Rational = Class{}
Rational:include(Unit)

function Rational:init(args)
  args.title = "Rational VCA"
  args.mnemonic = "RV"
  Unit.init(self,args)
end

function Rational:onLoadGraph(channelCount)
  if channelCount==2 then
    self:loadStereoGraph()
  else
    self:loadMonoGraph()
  end
end

function Rational:loadMonoGraph()
  local vca = self:createObject("RationalMultiply","vca",true)
  local numerator = self:createObject("GainBias","numerator")
  local numeratorRange = self:createObject("MinMax","numeratorRange")
  local divisor = self:createObject("GainBias","divisor")
  local divisorRange = self:createObject("MinMax","divisorRange")

  connect(numerator,"Out",numeratorRange,"In")
  connect(numerator,"Out",vca,"Numerator")
  connect(divisor,"Out",divisorRange,"In")
  connect(divisor,"Out",vca,"Divisor")

  connect(self,"In1",vca,"In")
  connect(vca,"Out",self,"Out1")

  self:createMonoBranch("numerator",numerator,"In",numerator,"Out")
  self:createMonoBranch("divisor",divisor,"In",divisor,"Out")
end

function Rational:loadStereoGraph()
  local vca1 = self:createObject("RationalMultiply","vca1",true)
  local vca2 = self:createObject("RationalMultiply","vca2",true)

  local numeratorL = self:createObject("GainBias","numeratorL")
  local numeratorRangeL = self:createObject("MinMax","numeratorRangeL")
  local divisorL = self:createObject("GainBias","divisorL")
  local divisorRangeL = self:createObject("MinMax","divisorRangeL")

  local numeratorR = self:createObject("GainBias","numeratorR")
  local numeratorRangeR = self:createObject("MinMax","numeratorRangeR")
  local divisorR = self:createObject("GainBias","divisorR")
  local divisorRangeR = self:createObject("MinMax","divisorRangeR")

  connect(numeratorL,"Out",numeratorRangeL,"In")
  connect(numeratorL,"Out",vca1,"Numerator")
  connect(divisorL,"Out",divisorRangeL,"In")
  connect(divisorL,"Out",vca1,"Divisor")

  connect(numeratorR,"Out",numeratorRangeR,"In")
  connect(numeratorR,"Out",vca2,"Numerator")
  connect(divisorR,"Out",divisorRangeR,"In")
  connect(divisorR,"Out",vca2,"Divisor")

  connect(self,"In1",vca1,"In")
  connect(self,"In2",vca2,"In")
  connect(vca1,"Out",self,"Out1")
  connect(vca2,"Out",self,"Out2")

  self:createMonoBranch("numeratorL",numeratorL,"In",numeratorL,"Out")
  self:createMonoBranch("divisorL",divisorL,"In",divisorL,"Out")
  self:createMonoBranch("numeratorR",numeratorR,"In",numeratorR,"Out")
  self:createMonoBranch("divisorR",divisorR,"In",divisorR,"Out")
end

function Rational:onLoadViews(objects,branches)
  local controls = {}
  local views = {
    collapsed = {},
  }

  if self.channelCount==2 then
    views.expanded = {"leftNum","leftDiv","rightNum","rightDiv"}

    controls.leftNum = GainBias {
      button = "num(L)",
      branch = branches.numeratorL,
      description = "Left Numerator",
      gainbias = objects.numeratorL,
      range = objects.numeratorRangeL,
      biasMap = Encoder.getMap("int[0,32]"),
      gainMap = Encoder.getMap("[-20,20]"),
      initialBias = 1,
      biasPrecision = 0
    }

    controls.leftDiv = GainBias {
      button = "div(L)",
      branch = branches.divisorL,
      description = "Left Divisor",
      gainbias = objects.divisorL,
      range = objects.divisorRangeL,
      biasMap = Encoder.getMap("int[1,32]"),
      gainMap = Encoder.getMap("[-20,20]"),
      initialBias = 1,
      biasPrecision = 0
    }

    controls.rightNum = GainBias {
      button = "num(R)",
      branch = branches.numeratorR,
      description = "Right Numerator",
      gainbias = objects.numeratorR,
      range = objects.numeratorRangeR,
      biasMap = Encoder.getMap("int[0,32]"),
      gainMap = Encoder.getMap("[-20,20]"),
      initialBias = 1,
      biasPrecision = 0
    }

    controls.rightDiv = GainBias {
      button = "div(R)",
      branch = branches.divisorR,
      description = "Right Divisor",
      gainbias = objects.divisorR,
      range = objects.divisorRangeR,
      biasMap = Encoder.getMap("int[1,32]"),
      gainMap = Encoder.getMap("[-20,20]"),
      initialBias = 1,
      biasPrecision = 0
    }
  else
    views.expanded = {"num","div"}

    controls.num = GainBias {
      button = "num",
      branch = branches.numerator,
      description = "Numerator",
      gainbias = objects.numerator,
      range = objects.numeratorRange,
      biasMap = Encoder.getMap("int[0,32]"),
      gainMap = Encoder.getMap("[-20,20]"),
      initialBias = 1,
      biasPrecision = 0
    }

    controls.div = GainBias {
      button = "div",
      branch = branches.divisor,
      description = "Divisor",
      gainbias = objects.divisor,
      range = objects.divisorRange,
      biasMap = Encoder.getMap("int[1,32]"),
      gainMap = Encoder.getMap("[-20,20]"),
      initialBias = 1,
      biasPrecision = 0
    }
  end

  return controls, views
end

return Rational
